/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */


using System;
using BEPUphysics;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;

namespace BEPUphysicsDemos.SampleCode
{
    /// <summary>
    /// Spring that attempts to keep a target entity oriented in a given direction by applying corrective torques.
    /// </summary>
    public class UprightSpring : Updateable
    {
        /*
         * This class is a cleaned up version of the UprightConstraint present in v0.11.0 and previous versions.
         * It's been moved into the demos as an example (and to clean up the main library).
         * 
         * This could be relatively easily re-implemented as a SingleEntityConstraint.  The SingleEntityAngularMotor
         * can already be used to do something fairly similar, but this spring has angle limits as well.
         * 
         */


        private Vector3 localUpVector;

        private float maximumAngle;
        private float minimumAngle;
        private Vector3 worldUpVector;


        /// <summary>
        /// Constructs a constraint to keep entities upright.
        /// </summary>
        /// <param name="entity">Entity to try to keep upright.</param>
        /// <param name="upVector">Direction to try to orient the entity with.</param>
        /// <param name="minimumAngle">Minimum angle between the car's transformed up vector and the actual up vector
        /// before the constraint begins to apply forces.</param>
        /// <param name="maximumAngle">Maximum angle between the car's transformed up vector and the actual up vector
        /// past which the constraint 'gives up' and lets the entity tumble.</param>
        /// <param name="correctionFactor">Factor of the orientation error to apply in angular velocity each frame.</param>
        public UprightSpring(Entity entity, Vector3 upVector, float minimumAngle, float maximumAngle, float correctionFactor)
        {
            Entity = entity;
            UpVector = upVector;
            MinimumAngle = minimumAngle;
            MaximumAngle = maximumAngle;
            CorrectionFactor = correctionFactor;
        }

        /// <summary>
        /// Gets or sets the factor of the orientation error to apply in torque each frame.
        /// </summary>
        public float CorrectionFactor { get; set; }

        /// <summary>
        /// Gets or sets the entity to try to keep upright.
        /// </summary>
        public Entity Entity { get; set; }

        /// <summary>
        /// Gets or sets the up vector attached to the entity in its local space.
        /// </summary>
        public Vector3 LocalUpVector
        {
            get { return worldUpVector; }
            set
            {
                localUpVector = Vector3.Normalize(value);
                worldUpVector = Vector3.TransformNormal(localUpVector, Entity.OrientationMatrix);
            }
        }

        /// <summary>
        /// Gets or sets the maximum angle between the car's transformed up vector and the actual up vector
        /// past which the constraint 'gives up' and lets the entity tumble.
        /// </summary>
        public float MaximumAngle
        {
            get { return maximumAngle; }
            set
            {
                maximumAngle = MathHelper.Clamp(value, 0, MathHelper.Pi);
                if (maximumAngle < minimumAngle)
                    MinimumAngle = maximumAngle;
            }
        }

        /// <summary>
        /// Gets or sets the minimum angle between the car's transformed up vector and the actual up vector
        /// before the constraint begins to apply forces.
        /// </summary>
        public float MinimumAngle
        {
            get { return minimumAngle; }
            set
            {
                minimumAngle = MathHelper.Clamp(value, 0, MathHelper.Pi);
                if (minimumAngle > maximumAngle)
                    MaximumAngle = minimumAngle;
            }
        }

        /// <summary>
        /// Gets or sets the up vector attached to the entity in world space.
        /// </summary>
        public Vector3 UpVector
        {
            get { return worldUpVector; }
            set
            {
                worldUpVector = Vector3.Normalize(value);
                localUpVector = Vector3.TransformNormal(worldUpVector, Matrix.Transpose(Entity.OrientationMatrix));
            }
        }

        /// <summary>
        /// Updates the upright constraint.
        /// Called automatically by its owning space.
        /// </summary>
        /// <param name="dt">Time since last frame in simulation seconds.</param>
        public override void UpdateDuringForces(float dt)
        {
            worldUpVector = Vector3.TransformNormal(localUpVector, Entity.InternalOrientationMatrix);

            //Compute the axis and angle 
            Vector3 axis = Vector3.Cross(worldUpVector, Vector3.Up);
            var angle = (float) Math.Acos(Vector3.Dot(Vector3.Up, worldUpVector));

            if (angle > MinimumAngle && angle < MaximumAngle)
            {
                angle = angle - MinimumAngle;
                axis.Normalize();
                Entity.ApplyAngularImpulse(axis * (angle * CorrectionFactor * dt));
            }
        }
    }
}